//============================================================================
//  aPCI-P31A for PCI Device Driver Diagnosis program
//  apci31a_diag.c
//  Author  Shoichiro Ito(ADTEK SYSTEM SCIENCE Co.,Ltd.)
//  Date    2001/03/15
//  Rev.    1.0.0
//  Note
//============================================================================

#include <stdio.h>	// used printf,perror...
#include <unistd.h>	// used read
#include <sys/ioctl.h>
#include <term.h>
#include <string.h>	// used strerror
#include <errno.h>	// used errno

#include "apci31actl.h"


// input of command from user
static char line[256];


//--- Ǵؿ ----------------------------------------------

static void diag01_create(void);
static void diag02_close(void);
static void diag03_inp(void);
static void diag04_outp(void);
static void diag05_get_ifr(void);
static void diag06_get_icr(void);
static void diag07_set_icr(void);
static void diag08_setup_int(void);
static void diag09_cancel_int(void);

static void diag10_test_inp(void);
static void diag11_test_outp(void);
static void diag12_test_int(void);

static void isr_proc(int bsn, void* arg);

static int    enter_bsn(void);
static struct termio init_termio(void);
static void   term_termio(struct termio tty);
static char*  itobina(int data, int size, char* dst);


//--- ؿơ֥ ------------------------------------------

static void (*pdiag_func[])(void) =
{
	diag01_create,
	diag02_close,
	diag03_inp,
	diag04_outp,
	diag05_get_ifr,
	diag06_get_icr,
	diag07_set_icr,
	diag08_setup_int,
	diag09_cancel_int,
	diag10_test_inp,
	diag11_test_outp,
	diag12_test_int
};

int main(int argc, char *argv[])
{
	int cmd, i;

	printf("aPCI-P31A diagnostic utility.\n");

	do
	{
		printf("\n");
		printf("aPCI-P31A Test main menu\n");
		printf("------------------------\n");
		printf(" Function test          \n");	// ǽƥ
		printf("------------------------\n");
		printf(" 1. apci31a_create(int bsn)                                \n");
		printf(" 2. apci31a_close(int bsn)                                 \n");
		printf(" 3. apci31a_inp(int bsn, int port, unsigned char* pdata)   \n");
		printf(" 4. apci31a_outp(int bsn, int port, unsigned char data)    \n");
		printf(" 5. apci31a_get_ifr(int bsn, unsigned char* pifr)          \n");
		printf(" 6. apci31a_get_icr(int bsn, unsigned short* picr)         \n");
		printf(" 7. apci31a_set_icr(int bsn, unsigned short icr)           \n");
		printf(" 8. apci31a_setup_int(int bsn, void (*pcallback)(int, void*), void* arg)\n");
		printf(" 9. apci31a_cancel_int(int bsn)                            \n");
		printf("------------------------\n");
		printf(" Application test       \n");	// ѥƥ
		printf("------------------------\n");
		printf("10. Inport                                                 \n");
		printf("11. Outport                                                \n");
		printf("12. Interrupt                                              \n");
		printf("------------------------\n");
		printf("99. Exit (All devices are closed)\n");
		printf("Enter option: ");

		cmd = 0;
		fgets(line, sizeof(line), stdin);
		sscanf(line, "%d", &cmd);

		if (cmd > 0 && cmd <= sizeof(pdiag_func) / sizeof(void*))
			pdiag_func[cmd - 1]();

		if(cmd != 99)
		{
			printf("--- Please enter key push! ---");
			fgets(line, sizeof(line), stdin);
		}
	}
	while (cmd != 99);

	// ƤΥǥХ򥯥
	for (i = 0; i < APCI31A_MAX_DEVICES; i++)
		apci31a_close(i);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  ǥХΥץ							    */
/*--------------------------------------------------------------------------*/
static void diag01_create(void)
{
	int bsn;

	printf("--- apci31a_create(int bsn)\n");

	// Get BSN
	printf("Enter bsn(0-15 or -1): ");
	fgets(line, sizeof(line), stdin);
	sscanf(line, "%d", &bsn);

	// Call Control Function
	bsn = apci31a_create(bsn);

	// result
	if (bsn < 0)
		printf("Rslt: %d\n", bsn);
	else
		printf("BSN: %d\n", bsn);
}

/*--------------------------------------------------------------------------*/
/*  ǥХΥ							    */
/*--------------------------------------------------------------------------*/
static void diag02_close(void)
{
	int bsn;
	int rslt;

	printf("--- apci31a_close(int bsn)\n");

	// Get BSN
	bsn = enter_bsn();

	// Call Control Function
	rslt = apci31a_close(bsn);

	printf("Rslt: %d\n", rslt);
}

/*--------------------------------------------------------------------------*/
/*  ϥݡȤǡ						    */
/*--------------------------------------------------------------------------*/
static void diag03_inp(void)
{
	int bsn, port;
	unsigned char data;
	int rslt;

	printf("--- apci31a_inp(int bsn, int port, unsigned char* pdata)\n");

	// Get BSN
	bsn = enter_bsn();

	// Get Port
	printf("Enter port: ");
	fgets(line, sizeof(line), stdin);
	sscanf(line, "%d", &port);

	// Call Control Function
	rslt = apci31a_inp(bsn, port, &data);

	printf("Rslt: %d\n", rslt);
	printf("data(hex): %x\n", data);
}

/*--------------------------------------------------------------------------*/
/*  ϥݡȤإǡ						    */
/*--------------------------------------------------------------------------*/
static void diag04_outp(void)
{
	int bsn, port;
	unsigned int data;
	int rslt;

	printf("--- apci31a_outp(int bsn, int port, unsigned char data)\n");

	// Get BSN
	bsn = enter_bsn();

	// Get Port
	printf("Enter port: ");
	fgets(line, sizeof(line), stdin);
	sscanf(line, "%d", &port);

	// Get Data
	printf("Enter data(hex): ");
	fgets(line, sizeof(line), stdin);
	sscanf(line, "%x", &data);

	// Call Control Function
	rslt = apci31a_outp(bsn, port, (unsigned char)data);

	printf("Rslt: %d\n", rslt);
}

/*--------------------------------------------------------------------------*/
/*  װ							    */
/*--------------------------------------------------------------------------*/
static void diag05_get_ifr(void)
{
	int bsn;
	unsigned char ifr;
	int rslt;

	printf("--- apci31a_get_ifr(int bsn, unsigned char* pifr)\n");

	// Get BSN
	bsn = enter_bsn();

	// Call Control Function
	rslt = apci31a_get_ifr(bsn, &ifr);

	printf("Rslt: %d\n", rslt);
	printf("ifr(hex): %x\n", ifr);
}

/*--------------------------------------------------------------------------*/
/*  ߾							    */
/*--------------------------------------------------------------------------*/
static void diag06_get_icr(void)
{
	int bsn;
	unsigned short icr;
	int rslt;

	printf("--- apci31a_get_icr(int bsn, unsigned short* picr)\n");

	// Get BSN
	bsn = enter_bsn();

	// Call Control Function
	rslt = apci31a_get_icr(bsn, &icr);

	printf("Rslt: %d\n", rslt);
	printf("icr(hex): %x\n", icr);
}

/*--------------------------------------------------------------------------*/
/*  ߾							    */
/*--------------------------------------------------------------------------*/
static void diag07_set_icr(void)
{
	int bsn;
	unsigned int icr;
	int rslt;

	printf("--- apci31a_set_icr(int bsn, unsigned short icr)\n");

	// Get BSN
	bsn = enter_bsn();

	// Get Interrupt Control Data
	printf("Enter icr(hex): ");
	fgets(line, sizeof(line), stdin);
	sscanf(line, "%x", &icr);

	// Call Control Function
	rslt = apci31a_set_icr(bsn, (unsigned short)icr);

	printf("Rslt: %d\n", rslt);
}

/*--------------------------------------------------------------------------*/
/*  ߥХå롼					    */
/*--------------------------------------------------------------------------*/
static void diag08_setup_int(void)
{
	int bsn;
	int rslt;

	printf("--- apci31a_setup_int(int bsn, void (*pcallback)(int, void*), void* arg)\n");

	// Get BSN
	bsn = enter_bsn();

	printf("\"pcallback\" and \"arg\" has been set up inside the program.\n\n");

	// Call Control Function
	rslt = apci31a_setup_int(bsn, isr_proc, NULL);

	printf("Rslt: %d\n", rslt);
}

/*--------------------------------------------------------------------------*/
/*  ߥХå롼μä				    */
/*--------------------------------------------------------------------------*/
static void diag09_cancel_int(void)
{
	int bsn;
	int rslt;

	printf("--- apci31a_cancel_int(int bsn)\n");

	// Get BSN
	bsn = enter_bsn();

	rslt = apci31a_cancel_int(bsn);

	printf("Rslt: %d\n", rslt);
}

/*--------------------------------------------------------------------------*/
/*  ϥݡȤǡϢ³					    */
/*--------------------------------------------------------------------------*/
static void diag10_test_inp(void)
{
	int    bsn, i;
	char   c, n;
	struct termio tty_save;
	char   data, szbuf[16];
	int    rslt;

	printf("--- The continuous test of Inport\n");

	// Get BSN
	bsn = enter_bsn();

	printf("\nA completion key is \"ESC\".\n\n");

	for (i = 0; i < APCI31A_MAX_PORTS; i++)
		printf( " CH%d IN   ", i);
	printf("\n");

	for (i = 0; i < APCI31A_MAX_PORTS; i++)
		printf("----------");
	printf("\n");

	tty_save = init_termio();

	for (;;)
	{
		c = '\0';		// ʤ硢ƤѤǤޤ
		n = read(0, &c, 1);
		if (n == -1)
		{
			perror("read");
			break;
		}
		else if (c == 0x1B)	// ESC key
			break;

		// InPort 
		for (i = 0; i < APCI31A_MAX_PORTS; i++)
		{
			rslt = apci31a_inp(bsn, i, &data);
			if (rslt < 0)
			{
				printf(" Rslt: %d ", rslt);
				continue;
			}

			printf(" %s ", itobina(data, sizeof(data), szbuf));
		}

		printf("\r");
		fflush(stdout);
	}

	// ᤹
	term_termio(tty_save);

	printf("\n");
}

/*--------------------------------------------------------------------------*/
/*  ϥݡȤإǡϢ³					    */
/*--------------------------------------------------------------------------*/
static void diag11_test_outp(void)
{
	int	  bsn, i;
	char	  c, n;
	struct	  termio tty_save;
	char	  data, szbuf[16];
	int	  rslt;

	printf("--- The continuous test of Outport\n");

	// Get BSN
	bsn = enter_bsn();

	printf("\nA completion key is \"ESC\".\n\n");

	for (i = 0; i < APCI31A_MAX_PORTS; i++)
		printf( " CH%d OUT  ", i);
	printf("\n");
	for (i = 0; i < APCI31A_MAX_PORTS; i++)
		printf("----------");
	printf("\n");

	tty_save = init_termio();

	data = 0x00;

	for (;;)
	{
		c = '\0';		// ʤ硢ƤѤǤޤ
		n = read(0, &c, 1);
		if (n == -1)
		{
			perror("read");
			break;
		}
		else if (c == 0x1B)	// ESC key
			break;

		// OutPort 
		for (i = 0; i < APCI31A_MAX_PORTS; i++)
		{
			rslt = apci31a_outp(bsn, i, data++);
			if (rslt < 0)
			{
				printf(" Rslt: %d ", rslt);
				continue;
			}

			printf(" %s ", itobina(data, sizeof(data), szbuf));
		}

		data++; // Dummy increment

		printf("\r");
		fflush(stdout);
	}

	// ᤹
	term_termio(tty_save);

	printf("\n");
}

/*--------------------------------------------------------------------------*/
/*  ߤϢ³							    */
/*--------------------------------------------------------------------------*/
static void diag12_test_int(void)
{
	int	 bsn, i;
	char	 c, n;
	struct	 termio tty_save;
	unsigned int icr;
	int	 rslt;
	int	 cnt_intr[APCI31A_MAX_IRQS];

	printf("--- The continuous test of Interrupt\n");

	// Get BSN
	bsn = enter_bsn();

	// Get Interrupt Control Data
	printf("Enter icr(hex): ");
	fgets(line, sizeof(line), stdin);
	sscanf(line, "%x", &icr);

	// Call Control Function
	rslt = apci31a_set_icr(bsn, (unsigned short)icr);
	if (rslt < 0)
	{
		printf("Rslt: %d\n", rslt);
		return;
	}

	for (i = 0; i < APCI31A_MAX_IRQS; i++)
		cnt_intr[i] = 0;	    // ߲ꥢ

	// Call Control Function
	rslt = apci31a_setup_int(bsn, isr_proc, cnt_intr);
	if (rslt < 0)
	{
		printf("Rslt: %d\n", rslt);
		return;
	}

	printf("\nA completion key is \"ESC\".\n\n");

	for (i = 0; i < APCI31A_MAX_IRQS; i++)
		printf("  IRQ%d ", i);	    // ɽ

	printf("\n----------------------------\n");

	tty_save = init_termio();

	for (;;)
	{
		c = '\0';		// ʤ硢ƤѤǤޤ
		n = read(0, &c, 1);
		if (n == -1)
		{
			perror("read");
			break;
		}
		else if (c == 0x1B)	// ESC key
			break;

		// ߥɽ
		for (i = 0; i < APCI31A_MAX_IRQS; i++)
			printf(" %5d ", cnt_intr[i]);

		printf("\r");
		fflush(stdout);
	}

	// ߶ػ
	rslt = apci31a_set_icr(bsn, (unsigned short)0x0000);
	if (rslt < 0)
		printf("Rslt: %d\n", rslt);

	// ᤹
	term_termio(tty_save);

	// Call Control Function
	rslt = apci31a_cancel_int(bsn);
	if (rslt < 0)
		printf("Rslt: %d\n", rslt);

	printf("\n");
}

/*--------------------------------------------------------------------------*/
/*  ߥХå롼					    */
/*--------------------------------------------------------------------------*/
static void isr_proc(int bsn, void* arg)
{
	unsigned char ifr;
	int rslt;
	int* cnt_intr;

	rslt = apci31a_get_ifr(bsn, &ifr);
	if (rslt < 0)
	{
		printf("Rslt(ifr): %d\n", rslt);
		return;
	}

	if (arg == NULL)
		return;

	cnt_intr = (int*)arg;

	if (ifr & 0x01) cnt_intr[0]++;
	if (ifr & 0x02) cnt_intr[1]++;
	if (ifr & 0x04) cnt_intr[2]++;
	if (ifr & 0x08) cnt_intr[3]++;
}

/*--------------------------------------------------------------------------*/
/*  BSN 								    */
/*--------------------------------------------------------------------------*/
static int enter_bsn(void)
{
	int bsn;

	printf("Enter bsn: ");
	fgets(line, sizeof(line), stdin);
	sscanf(line, "%d", &bsn);

	return bsn;
}

/*--------------------------------------------------------------------------*/
/*  ꥢ륿७						    */
/*--------------------------------------------------------------------------*/
static struct termio init_termio(void)
{
	struct termio tty, tty_save;

	// ߤ¸
	ioctl(0, TCGETA, &tty);

	tty_save = tty;

	// ü
	tty.c_lflag &= ~(ECHO | ICANON);
	tty.c_cc[VMIN] = 0;
	tty.c_cc[VTIME] = 2;
	ioctl(0, TCSETAF, &tty);

	return tty_save;
}

/*--------------------------------------------------------------------------*/
/*  ꥢ륿७ϲ						    */
/*--------------------------------------------------------------------------*/
static void term_termio(struct termio tty)
{
	// ᤹
	ioctl (0, TCSETAF, &tty);
}

/*--------------------------------------------------------------------------*/
/*   ---> 2 ʿʸ						    */
/*--------------------------------------------------------------------------*/
static char* itobina(int data, int size, char* dst)
{
	int i;
	int bits = size * 8;
	for (i = 0; i < bits; i++, data >>= 1)
		dst[bits - 1 - i] = '0' + (data & 0x1);
	dst[bits] = '\0';

	return dst;
}
