//============================================================================
//  aPCI-P31A Device Driver I/F Control
//  apci31actl.c
//  Author  Shoichiro Ito(ADTEK SYSTEM SCIENCE Co.,Ltd.)
//  Date    2001/03/15
//  Rev.    1.0.0
//============================================================================

#include <stdio.h>	// used printf,perror...
#include <pthread.h>	// used pthread
#include <fcntl.h>	// used open
#include <unistd.h>	// used close,usleep
#include <sys/ioctl.h>	// used ioctl
#include <sys/poll.h>	// used poll
#include <string.h>	// used strerror
#include <errno.h>	// used errno

#include "apci31actl.h"


/*----------------- Define Device Control ----------------------------------*/

#define APCI31A_FNC_EXEC	1		// ¹
#define APCI31A_FNC_NOTEXEC	0		// ¹Բǽ

typedef struct _apci31a_ctl
{
	int	  fd;
	pthread_t hisr;
	void	  (*pcallback)(int, void*);
	void*	  arg;
	int	  exit_isr;
	int	  bexec;			// ǽ¹
} apci31a_ctl, *papci31a_ctl;

static apci31a_ctl ctl[APCI31A_MAX_DEVICES] =
{
	{ 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	{ 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }
};


/*----------------- Define Static Functions --------------------------------*/

static int	dev_open(int bsn);
static int	is_bsn(int bsn);
static int	is_pass(int bsn);
static void*	thread_isr(void* arg);


/*==========================================================================*/
/*  apci31a_create - ǥХץ					    */
/*  Input : bsn - BSN (APCI31A_BSN_AUTO ξ硢̤ѥǥХ)	    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 ʾ BSN 					    */
/*==========================================================================*/
int apci31a_create(int bsn)
{
	int fd = 0, i;

	// BSN 
	if (bsn != APCI31A_BSN_AUTO)
	{
		// BSN 
		if (!(bsn >= 0 && bsn < APCI31A_MAX_DEVICES))
			return APCI31A_EINVAL;

		// ǥХץ
		fd = dev_open(bsn);
		if (fd < 0)
		{
			PDEBUG("Error: open, %s\n", strerror(errno));

			if (errno == EBUSY)
				return APCI31A_EBUSY;
			return APCI31A_ENODEV;
		}

		ctl[bsn].fd    = fd;
		ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;
		ctl[bsn].hisr  = 0;

		return bsn;
	}

	// ̤ѥǥХץ
	for (i = 0; i < APCI31A_MAX_DEVICES; i++)
	{
		// ǥХϥץ
		if (ctl[i].fd != 0)
			continue;

		// ǥХץ
		fd = dev_open(i);
		if (fd < 0)
			continue;

		break;
	}

	// ץǤǥХ̵
	if (i >= APCI31A_MAX_DEVICES)
		return APCI31A_ENODEV;

	ctl[i].fd    = fd;
	ctl[i].bexec = APCI31A_FNC_NOTEXEC;
	ctl[i].hisr  = 0;

	return i;		// return BSN
}

/*==========================================================================*/
/*  apci31a_close - ǥХ					    */
/*  Input : bsn - BSN 						    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci31a_close(int bsn)
{
	int rslt;

	// åɤƤϽλ
	rslt = apci31a_cancel_int(bsn);
	if (rslt != APCI31A_OK)
		return rslt;

	// ѥ᡼
	// BSN 
	if (!is_bsn(bsn))
		return APCI31A_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(bsn))
		return APCI31A_EBUSY;

	ctl[bsn].bexec = APCI31A_FNC_EXEC;		// ¹

	// ǥХΥ
	close(ctl[bsn].fd);

	ctl[bsn].fd    = 0;
	ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;		// ¹Բǽ

	return APCI31A_OK;
}

/*==========================================================================*/
/*  apci31a_inp - ϥݡȤǡ				    */
/*  Input : bsn	  - BSN 						    */
/*	    port  - ݡȻ						    */
/*	    pdata - ǡǼΰ					    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci31a_inp(int bsn, int port, unsigned char* pdata)
{
	apci31a_io iobuf;

	// ѥ᡼
	// BSN 
	if (!is_bsn(bsn))
		return APCI31A_EINVAL;

	if (pdata == NULL)
		return APCI31A_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(bsn))
		return APCI31A_EBUSY;

	ctl[bsn].bexec = APCI31A_FNC_EXEC;		// ¹

	// ѥ᡼
	iobuf.port = port;

	// ioctl ¹
	if (ioctl(ctl[bsn].fd, APCI31A_IN_PORT, &iobuf) < 0)
	{
		PDEBUG("Error: ioctl, %s\n", strerror(errno));

		ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;	// ¹Բǽ
		return APCI31A_ESYSCALL;
	}

	*pdata = iobuf.value;

	ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;		// ¹Բǽ

	return APCI31A_OK;
}

/*==========================================================================*/
/*  apci31a_outp - ϥݡȤإǡ				    */
/*  Input : bsn	 - BSN 						    */
/*	    port - ݡȻ						    */
/*	    data - ϥǡ						    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci31a_outp(int bsn, int port, unsigned char data)
{
	apci31a_io iobuf;

	// ѥ᡼
	// BSN 
	if (!is_bsn(bsn))
		return APCI31A_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(bsn))
		return APCI31A_EBUSY;

	ctl[bsn].bexec = APCI31A_FNC_EXEC;		// ¹

	// ѥ᡼
	iobuf.port  = port;
	iobuf.value = data;

	// ioctl ¹
	if (ioctl(ctl[bsn].fd, APCI31A_OUT_PORT, &iobuf) < 0)
	{
		PDEBUG("Error: ioctl, %s\n", strerror(errno));

		ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;	// ¹Բǽ
		return APCI31A_ESYSCALL;
	}

	ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;		// ¹Բǽ

	return APCI31A_OK;
}

/*==========================================================================*/
/*  apci31a_get_ifr - װ				    */
/*  Input : bsn	 - BSN 						    */
/*	    pifr - װǼΰ					    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci31a_get_ifr(int bsn, unsigned char* pifr)
{
	// ѥ᡼
	// BSN 
	if (!is_bsn(bsn))
		return APCI31A_EINVAL;

	if (pifr == NULL)
		return APCI31A_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(bsn))
		return APCI31A_EBUSY;

	ctl[bsn].bexec = APCI31A_FNC_EXEC;		// ¹

	// ioctl ¹
	if (ioctl(ctl[bsn].fd, APCI31A_GET_IFR, pifr) < 0)
	{
		PDEBUG("Error: ioctl, %s\n", strerror(errno));

		ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;	// ¹Բǽ
		return APCI31A_ESYSCALL;
	}

	ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;		// ¹Բǽ

	return APCI31A_OK;
}

/*==========================================================================*/
/*  apci31a_get_icr - ߾					    */
/*  Input : bsn	 - BSN 						    */
/*	    picr - ߾Ǽΰ					    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci31a_get_icr(int bsn, unsigned short* picr)
{
	// ѥ᡼
	// BSN 
	if (!is_bsn(bsn))
		return APCI31A_EINVAL;

	if (picr == NULL)
		return APCI31A_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(bsn))
		return APCI31A_EBUSY;

	ctl[bsn].bexec = APCI31A_FNC_EXEC;		// ¹

	// ioctl ¹
	if (ioctl(ctl[bsn].fd, APCI31A_GET_ICR, picr) < 0)
	{
		PDEBUG("Error: ioctl, %s\n", strerror(errno));

		ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;	// ¹Բǽ
		return APCI31A_ESYSCALL;
	}

	ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;		// ¹Բǽ

	return APCI31A_OK;
}

/*==========================================================================*/
/*  apci31a_set_icr - ߾					    */
/*  Input : bsn - BSN 						    */
/*	    icr - ߾						    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci31a_set_icr(int bsn, unsigned short icr)
{
	// ѥ᡼
	// BSN 
	if (!is_bsn(bsn))
		return APCI31A_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(bsn))
		return APCI31A_EBUSY;

	ctl[bsn].bexec = APCI31A_FNC_EXEC;		// ¹

	// ioctl ¹
	if (ioctl(ctl[bsn].fd, APCI31A_SET_ICR, &icr) < 0)
	{
		PDEBUG("Error: ioctl, %s\n", strerror(errno));

		ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;	// ¹Բǽ
		return APCI31A_ESYSCALL;
	}

	ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;		// ¹Բǽ

	return APCI31A_OK;
}

/*==========================================================================*/
/*  apci31a_setup_int - ߥХå롼		    */
/*  Input : bsn	      - BSN 					    */
/*	    pcallback - ߥХå롼Υɥ쥹		    */
/*	    arg	      - Хå롼Ϥ桼(NULL )	    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci31a_setup_int(int bsn, void(*pcallback)(int bsn, void* arg), void* arg)
{
	int rslt;

	// ѥ᡼
	// BSN 
	if (!is_bsn(bsn))
		return APCI31A_EINVAL;

	// Хåؿɥ쥹
	if (pcallback == NULL)
		return APCI31A_EINVAL;

	// ˥Хå롼ꤵƤ
	if (ctl[bsn].hisr != 0)
		return APCI31A_ESISR;			// 

	// ǽμ¹Բ
	if (!is_pass(bsn))
		return APCI31A_EBUSY;

	ctl[bsn].bexec = APCI31A_FNC_EXEC;		// ¹

	ctl[bsn].pcallback = pcallback;			// Хå롼
	ctl[bsn].arg	   = arg;			// 桼
	ctl[bsn].exit_isr  = 0;				// åɽλ

	rslt = pthread_create(&ctl[bsn].hisr, NULL, thread_isr, (void*)bsn);
	if (rslt)
	{
		PDEBUG("Error: pthread_create, %s\n", strerror(rslt));

		ctl[bsn].hisr  = 0;			// ǰΤ˽

		ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;	// ¹Բǽ
		return APCI31A_ESISR;			// 꼺
	}

	ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;		// ¹Բǽ

	return APCI31A_OK;
}

/*==========================================================================*/
/*  apci31a_cancel_int - ߥХå롼μä		    */
/*  Input : bsn - BSN 						    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci31a_cancel_int(int bsn)
{
	// ѥ᡼
	// BSN 
	if (!is_bsn(bsn))
		return APCI31A_EINVAL;

	// Хå롼ꤵƤʤ
	if (ctl[bsn].hisr == 0)
		return APCI31A_OK;			// ｪλ

	// ǽμ¹Բ
	if (!is_pass(bsn))
		return APCI31A_EBUSY;

	ctl[bsn].bexec = APCI31A_FNC_EXEC;		// ¹

	ctl[bsn].exit_isr = 1;				// åɽλ
	pthread_cancel(ctl[bsn].hisr);

	ctl[bsn].hisr = 0;				// åɥϥɥ

	ctl[bsn].bexec = APCI31A_FNC_NOTEXEC;		// ¹Բǽ

	return APCI31A_OK;
}

/*--------------------------------------------------------------------------*/
/*  dev_open - ǥХΥץ					    */
/*  Input : bsn - BSN 						    */
/*  Output: return - open 					    */
/*--------------------------------------------------------------------------*/
static int dev_open(int bsn)
{
	char szbuf[30];

	sprintf(szbuf, "/dev/apci31a%d", bsn);

	return open(szbuf, O_RDWR);
}

/*--------------------------------------------------------------------------*/
/*  is_bsn - BSN å						    */
/*  Input : bsn - BSN 						    */
/*  Output: return - 0 ̵						    */
/*		     1 						    */
/*--------------------------------------------------------------------------*/
static int is_bsn(int bsn)
{
	// ϰϤ
	if (!(bsn >= 0 && bsn < APCI31A_MAX_DEVICES))
		return 0;

	// ǥХξ
	if (ctl[bsn].fd == 0)
		return 0;

	// Ѳǽ
	return 1;
}

/*--------------------------------------------------------------------------*/
/*  is_pass - ¹Բǽå						    */
/*  Input : bsn - BSN 						    */
/*  Output: return - 0 ̵						    */
/*		     1 						    */
/*--------------------------------------------------------------------------*/
static int is_pass(int bsn)
{
	// ǽμ¹Ծ
	if (ctl[bsn].bexec != APCI31A_FNC_NOTEXEC)
		return 0;	// ¹Բ

	// ¹Բǽ
	return 1;
}

/*--------------------------------------------------------------------------*/
/*  thread_isr - ߽å					    */
/*  Input : arg - BSN 						    */
/*  Output: return - (void*)0						    */
/*--------------------------------------------------------------------------*/
static void* thread_isr(void* arg)
{
	struct pollfd ufds;

	int bsn = (int)arg;
	int rslt;

	for (;;)
	{
		// λ郎Ω硢ʲνϤʤ褦ˤ롣
		if (ctl[bsn].exit_isr)
			break;

		// ߼Ե
		// poll ΣܤΰϡǤ
		// ܤΰϡ-1(ॢ̵)Ǥ
		ufds.fd	    = ctl[bsn].fd;
		ufds.events = POLLIN;
		rslt = poll(&ufds, 1, -1);
		if (rslt < 0)
		{
			PDEBUG("Error: poll, %s\n", strerror(rslt));
			continue;
		}

		// 󥹥ȡ뤵줿Хå롼θƤӽФ
		ctl[bsn].pcallback(bsn, ctl[bsn].arg);
	}

	return (void*)0;
}
